# tab_power.py

import tkinter as tk
from tkinter import ttk
from tkinter.scrolledtext import ScrolledText
import math

class TabPower:
    def __init__(self, notebook, tab_coil, tab_simulator):
        """
        Pestaña para dimensionar C o V0 a partir de la energía mecánica deseada,
        considerando n etapas/bobinas idénticas.
        
        :param notebook: ttk.Notebook donde se añade esta pestaña
        :param tab_coil: referencia a la instancia de TabCoil
        :param tab_simulator: referencia a la instancia de TabSimulator (para leer E_target)
        """
        self.notebook = notebook
        self.tab_coil = tab_coil
        self.tab_simulator = tab_simulator
        
        # Frame principal para la pestaña
        self.frame = tk.Frame(notebook)
        self.frame.pack(fill="both", expand=True)
        
        # Añadir la pestaña al notebook
        notebook.add(self.frame, text="Power Stages (Energy-driven)")

        # Layout: parte izquierda (entradas) y derecha (resultados / logs)
        self.frame_left = tk.Frame(self.frame)
        self.frame_left.pack(side="left", fill="y", padx=5, pady=5)

        self.frame_right = tk.Frame(self.frame, bd=2, relief="groove")
        self.frame_right.pack(side="right", fill="both", expand=True, padx=5, pady=5)

        # ------------------------------
        # Variables y widgets de entrada
        # ------------------------------
        # Modo de dimensionado
        self.mode_var = tk.StringVar(value="Dimensionar C")  # o "Dimensionar V0"
        tk.Label(self.frame_left, text="Modo de Dimensionado:").pack(anchor="w")
        self.combo_mode = ttk.Combobox(
            self.frame_left, textvariable=self.mode_var,
            values=["Dimensionar C", "Dimensionar V0"],
            state="readonly"
        )
        self.combo_mode.bind("<<ComboboxSelected>>", self.on_mode_change)
        self.combo_mode.pack(pady=2)

        # Número de etapas
        tk.Label(self.frame_left, text="Número de etapas (n):").pack(anchor="w")
        self.var_n = tk.IntVar(value=3)
        tk.Entry(self.frame_left, textvariable=self.var_n, width=10).pack(pady=2)

        # Capacidad (si se dimensiona V0, la capacit se fija)
        tk.Label(self.frame_left, text="Capacidad (µF):").pack(anchor="w")
        self.var_C = tk.DoubleVar(value=100.0)  # ejemplo
        self.entry_C = tk.Entry(self.frame_left, textvariable=self.var_C, width=10)
        self.entry_C.pack(pady=2)

        # Tensión inicial (si se dimensiona C, la tensión se fija)
        tk.Label(self.frame_left, text="Tensión inicial (V):").pack(anchor="w")
        self.var_V0 = tk.DoubleVar(value=200.0)  # ejemplo
        self.entry_V0 = tk.Entry(self.frame_left, textvariable=self.var_V0, width=10)
        self.entry_V0.pack(pady=2)

        # Tiempo de descarga aproximado
        tk.Label(self.frame_left, text="Tiempo de etapa (ms):").pack(anchor="w")
        self.var_time_ms = tk.DoubleVar(value=10.0)  # 10 ms
        tk.Entry(self.frame_left, textvariable=self.var_time_ms, width=10).pack(pady=2)

        # Botón dimensionar
        tk.Button(self.frame_left, text="Dimensionar", command=self.dimensionar).pack(pady=5)

        # ------------------------------
        # Sección derecha: Logs/Resultados
        # ------------------------------
        tk.Label(self.frame_right, text="Resultados:", font=("Arial",10,"bold")).pack(anchor="nw", padx=5, pady=5)
        self.log_area = ScrolledText(self.frame_right, wrap="word", height=20, width=50)
        self.log_area.pack(fill="both", expand=True, padx=5, pady=5)
        self.log_area.configure(state="disabled")

        # Forzar actualización inicial de estados de input
        self.on_mode_change()

    def on_mode_change(self, event=None):
        """Habilita o deshabilita las entradas de C o V0 según el modo de dimensionado."""
        mode = self.mode_var.get()
        if mode == "Dimensionar C":
            # Bloqueamos el field de C para escritura manual?
            # O, según la lógica, convendría DESBLOQUEAR C si lo vamos a dimensionar?
            # Típicamente: dimensionar C => el usuario fija V0
            self.entry_C.config(state="disabled")  # Se calculará
            self.entry_V0.config(state="normal")   # Se ingresa manualmente
        else:
            # Dimensionar V0 => el usuario fija C
            self.entry_C.config(state="normal")
            self.entry_V0.config(state="disabled")

    def log_message(self, text, mode="info"):
        """Inserta texto en el área de log."""
        self.log_area.configure(state="normal")
        if mode=="error":
            self.log_area.insert("end", "ERROR: " + text + "\n", "error")
        else:
            self.log_area.insert("end", text + "\n", mode)
        self.log_area.see("end")
        self.log_area.configure(state="disabled")

    def dimensionar(self):
        """
        Lee la energía objetivo desde TabSimulator, y ajusta C o V0
        para que la energía total (n etapas) se aproxime a E_target.
        """
        # 0) Energía objetivo proveniente de TabSimulator
        E_target_total = self.tab_simulator.get_energy_required()
        if E_target_total <= 0:
            self.log_message("E_target <= 0. Asegúrate de que TabSimulator calcule algo válido.", "error")
            return

        # n (etapas)
        n = self.var_n.get()
        if n < 1:
            self.log_message("El número de etapas debe ser >= 1", "error")
            return

        #  => La energía por etapa
        E_target_stage = E_target_total / n

        # 1) Leer inductancia y resistencia desde TabCoil
        try:
            L_uH = float(self.tab_coil.var_L.get())  # microHenrios
            R_coil = float(self.tab_coil.var_R.get())  # Ohms
        except ValueError:
            self.log_message("No se pudo leer L/R de TabCoil.", "error")
            return
        L_coil = L_uH * 1e-6  # Henrios

        # h_c
        try:
            h_c_mm = float(self.tab_coil.var_h_c.get())
        except ValueError:
            self.log_message("No se pudo leer h_c de TabCoil.", "error")
            return
        h_c = h_c_mm / 1000.0  # m

        # 2) Leer los datos de (C, V0, dt) (aunque uno de los dos se va a dimensionar)
        dt_ms = self.var_time_ms.get()
        if dt_ms <= 0:
            self.log_message("Tiempo de etapa (ms) inválido", "error")
            return
        T_total = dt_ms*1e-3

        mode = self.mode_var.get()

        # Obtenemos un valor 'fijo' para el que no dimensionamos
        # y definimos la test_function acorde
        if mode=="Dimensionar C":
            # El user fija V0:
            try:
                V0_fixed = float(self.var_V0.get())
                if V0_fixed <= 0:
                    self.log_message("Tensión (V) inválida", "error")
                    return
            except ValueError:
                self.log_message("Tensión (V) inválida", "error")
                return

            # Rango de bisección en [Cmin, Cmax], µF
            lower = 0.1
            upper = 1e6
            def test_function(C_uF):
                E_stage = self.simular_descarga(
                    C_uF, V0_fixed, L_coil, R_coil, T_total, h_c
                )
                return E_stage

        else:
            # Dimensionar V0 => user fija C
            try:
                C_uF_fixed = float(self.var_C.get())
                if C_uF_fixed<=0:
                    self.log_message("Capacidad (µF) inválida", "error")
                    return
            except ValueError:
                self.log_message("Capacidad (µF) inválida", "error")
                return

            # Rango de bisección en [1..5000] V, p. ej.
            lower = 1.0
            upper = 5000.0
            def test_function(V0):
                E_stage = self.simular_descarga(
                    C_uF_fixed, V0, L_coil, R_coil, T_total, h_c
                )
                return E_stage

        # Bisección
        E_low = test_function(lower)
        E_up = test_function(upper)

        # Comprobación de que la energía buscada (E_target_stage) está entre E_low y E_up
        if not ((E_low <= E_target_stage <= E_up) or (E_up <= E_target_stage <= E_low)):
            self.log_message("ADVERTENCIA: El rango de bisección puede no encapsular la E_target. Ajusta límites.", "error")

        iteration_count = 0
        max_iter = 40
        tol = 0.01*E_target_stage  # 1% de la energía por etapa
        best_param = 0.0
        while iteration_count < max_iter:
            mid = 0.5*(lower + upper)
            E_mid = test_function(mid)
            diff = E_mid - E_target_stage
            if abs(diff) < tol:
                best_param = mid
                break
            # Suponemos monotonicidad: +C => +E, +V0 => +E
            if E_mid > E_target_stage:
                upper = mid
            else:
                lower = mid
            iteration_count += 1
            best_param = mid

        E_final_stage = test_function(best_param)
        E_final_total = E_final_stage * n
        diff_total = E_final_total - E_target_total

        # 3) Mostrar resultados
        self.log_message("-"*40)
        self.log_message(f"E_target_total = {E_target_total:.3f} J, n = {n}")
        self.log_message(f"E_target_stage = {E_target_stage:.3f} J", "info")

        if mode=="Dimensionar C":
            self.log_message(f"Dimensionado de C => V0={V0_fixed:.1f} V fijo.", "info")
            self.log_message(f"Capacidad final ~ {best_param:.2f} µF", "info")
        else:
            self.log_message(f"Dimensionado de V0 => C={self.var_C.get():.1f} µF fijo.", "info")
            self.log_message(f"Tensión final ~ {best_param:.2f} V", "info")

        self.log_message(f"E_result_stage ~ {E_final_stage:.3f} J => total ~ {E_final_total:.3f} J (diff={diff_total:.3f})", "info")
        self.log_message("-"*40)


    def simular_descarga(self, C_uF, V0, L_coil, R_coil, T_total, h_c):
        """
        Modelo simplificado R-L(x)-C de una 'etapa' para un tiempo T_total.
        Devuelve la energía consumida del condensador.
        
        Reutilizamos la lógica de Euler para la descarga, 
        asumiendo L varía ~ 20% a lo largo x in [0..h_c].
        """
        if C_uF<=0 or V0<=0 or T_total<=0:
            return 0.0

        # Convertimos C a F
        C = C_uF * 1e-6
        # Asumimos L variable ~ 20% de variación
        L0 = 1.2*L_coil
        Lf = 0.8*L_coil

        def L_of_x(x):
            return L0 + (Lf - L0)*(x/h_c) if h_c>1e-9 else L_coil
        def dLdx_of_x(x):
            return (Lf - L0)/h_c if h_c>1e-9 else 0.0

        N_steps = 1000
        dt = T_total/N_steps
        vCap = V0
        i = 0.0

        def x_of_t(t):
            return (h_c / T_total)*t if T_total>1e-9 else 0.0

        for step in range(N_steps+1):
            x = x_of_t(step*dt)
            Lx = L_of_x(x)
            dLdt = dLdx_of_x(x)*(h_c / T_total)

            if abs(Lx)>1e-12:
                di_dt = (vCap - R_coil*i - i*dLdt)/Lx
            else:
                di_dt = 0.0

            i_new = i + di_dt*dt
            vCap_new = vCap - (i/C)*dt
            if vCap_new<0:
                vCap_new=0.0
            i = i_new
            vCap = vCap_new

        # Energía consumida ~ 1/2*C*(V0^2 - vCap^2)
        E_consumida = 0.5*C*(V0**2 - vCap**2)
        if E_consumida<0:
            E_consumida=0.0
        return E_consumida